/**
 * 
 * @author Christian Rempis
 * c.rempis@gmx.de
 * 
 */
package de.uos.ikw.octavio.spinalcordmanager.core;

import java.awt.Color;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import de.uos.ikw.octavio.spinalcordmanager.formatter.DataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.EmptyDataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.Float8_8DataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.Int16DataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.Int8DataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.NormalizedDoubleFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.SyncDataFormatter;

public class DataManager implements SystemObject {

	/** Member Variables *********************************************************
	 *****************************************************************************/

	private Hashtable<String, DataFormatter> mAvailableDataFormatters = new Hashtable<String, DataFormatter>();
	private Hashtable<String, Color> mAvailableDataColors = new Hashtable<String, Color>();
	
	private Vector<DataListener> mDataListeners = new Vector<DataListener>();
	private LinkedList<DataFrame> mDataHistory = new LinkedList<DataFrame>();
	
	private Object[] mCurrentData;
	private DataFormatter[] mDataFormatter;
	private Color[] mDataColors;
	private int mNumberOfColumns;
	private int mNumberOfRows;
	private int mMaxHistorySize;
	private int mNumberOfBytesPerFrame;
	private boolean mLogDataToFile;
	private String mLogDataFileName;
	private FileWriter mDataFileWriter;
	
	private int mMasterSectionSize;
	private int mSlot0SectionSize;
	private int mSlotNSectionSize;
	private int mFrameNumberIndex;
	
	private boolean mAutoSetData;
	private int mLastReceivedFrameNumber;
	private int mLastInterpretedFrameNumber;
	
	private byte[] mLastInterpretedDataFrame;
	private boolean mIgnoreFrameSynchronization;
	

	/** Constructors *************************************************************
	 *****************************************************************************/
	
	
	public DataManager() {
		mCurrentData = new Object[0];
		mDataFormatter = new DataFormatter[0];
		mDataColors = new Color[0];
		mNumberOfColumns = 0;
		mNumberOfRows = 0;
		mMaxHistorySize = 5000;
		mNumberOfBytesPerFrame = 0;
		mFrameNumberIndex = 0;
		mLastReceivedFrameNumber = 0;
		mIgnoreFrameSynchronization = false;
		
		
		//Add default DataFormatters, so that they can be found for dynamic data interpretation.
		addAvailableDataFormatter(new SyncDataFormatter());
		addAvailableDataFormatter(new Int8DataFormatter());
		addAvailableDataFormatter(new Int16DataFormatter());
		addAvailableDataFormatter(new Float8_8DataFormatter());
		addAvailableDataFormatter(new EmptyDataFormatter("Empty0", 0));
		addAvailableDataFormatter(new EmptyDataFormatter("Empty1", 1));
		addAvailableDataFormatter(new EmptyDataFormatter("Empty2", 2));
		addAvailableDataFormatter(new EmptyDataFormatter("Empty4", 4));
		addAvailableDataFormatter(new EmptyDataFormatter("Empty8", 8));
		addAvailableDataFormatter(new NormalizedDoubleFormatter());
		
		
		//Add default Color objects for dynamic data visualization.
		mAvailableDataColors.put("red", Color.red);
		mAvailableDataColors.put("cyan", Color.cyan);
		mAvailableDataColors.put("blue", Color.blue);
		mAvailableDataColors.put("green", Color.green);
		mAvailableDataColors.put("orange", Color.orange);
		mAvailableDataColors.put("lightgray", Color.lightGray);
		mAvailableDataColors.put("gray", Color.gray);
		mAvailableDataColors.put("white", Color.white);
		mAvailableDataColors.put("yellow", Color.yellow);
		mAvailableDataColors.put("magenta", Color.magenta);
		mAvailableDataColors.put("pink", Color.pink);
		
		mAutoSetData = true;
	}


		
	/** Public Static ************************************************************
	 *****************************************************************************/
		
	
	


		
	/** Public *******************************************************************
	 *****************************************************************************/
		
		
	public boolean init() {
		return true;
	}

	public boolean shutDown() {
		return true;
	}	
	
	public boolean setLogFile(String fileName) {
		if(mDataFileWriter != null) {
			try {
				mDataFileWriter.close();
			} catch (IOException e) {
			}
			mDataFileWriter = null;
		}
		mLogDataFileName = fileName;
		if(mLogDataToFile) {
			try {
				mDataFileWriter = new FileWriter(mLogDataFileName);
			} catch (IOException e) {
				System.out.println("Could not open file [" + mLogDataFileName + "] for writing.");
				return false;
			}
		}
		return true;
	}
	
	public boolean setLogData(boolean logData) {
		mLogDataToFile = logData;
		if(mLogDataToFile) {
			try {
				mDataFileWriter = new FileWriter(mLogDataFileName);
			} catch (IOException e) {
				System.out.println("Could not open file [" + mLogDataFileName + "] for writing.");
				return false;
			}
		}
		else {
			if(mDataFileWriter != null) {
				try {
					mDataFileWriter.close();
				} catch (IOException e) {
				}
				mDataFileWriter = null;
			}
		}
		return true;
	}
	
//	public void notifyDataChanged() {
//		mModel.fireTableDataChanged();
//		mModel.fireTableStructureChanged();
//		if(mPaintCounter++ >= 100) {
//			mTable.repaint();
//			mPaintCounter = 0;
//		}
		
//		notifyDataContentChanged();
//	}
	
	
	public boolean setRawData(byte[] data) {
		
		if(data == null) {
			return false;
		}
		if(data.length != mNumberOfBytesPerFrame) {
			System.out.println("Number of bytes per frame does not fit: Required: " + mNumberOfBytesPerFrame + " but is " + data.length);
			return false;
		}
		if(mFrameNumberIndex >= 0 && mFrameNumberIndex < data.length - 1) {
			mLastInterpretedFrameNumber = (int) (((0xFF & data[mFrameNumberIndex + 1]) << 8) | (0xFF & data[mFrameNumberIndex]));
		}
		int dataCounter = 0;
		DataFormatter formatter = null;
		for(int i = 0; i < mDataFormatter.length; ++i) {
			formatter = mDataFormatter[i];
			if(formatter == null) {
				continue;
			}
			if(dataCounter >= data.length || mCurrentData.length <= i) {
				System.out.println("Counter larger than data size: " + i + " " + dataCounter + " " + data.length);
				break;
			}
			mCurrentData[i] = formatter.getFormattedData(data, dataCounter);
			dataCounter += formatter.getNumberOfBytesPerData();
		}
		
		
		if(mLogDataToFile && mDataFileWriter != null) {
			for(int i = 0; i < mCurrentData.length; ++i) {
				Object obj = mCurrentData[i];
				try {
					if(obj == null) {
						mDataFileWriter.write("null ");
					}
					else {
						mDataFileWriter.write(obj.toString() + " ");
					}
				} catch (IOException e) {
				}
			}
			try {
				mDataFileWriter.write("\n");
				mDataFileWriter.flush();
			} catch (IOException e) {
			}
		}
		
		mLastInterpretedDataFrame = data;
		
		notifyDataContentChanged();
		return true;
	}
	
	public byte[] getLastInterpretedDataFrame() {
		if(mLastInterpretedDataFrame == null) {
			int size = 0;
			for(int i = 0; i < mDataFormatter.length; ++i) {
				if(mDataFormatter[i] != null) {
					size += mDataFormatter[i].getNumberOfBytesPerData();
				}
			}
			mLastInterpretedDataFrame = new byte[size];
			for(int i = 0; i < mLastInterpretedDataFrame.length; ++i) {
				mLastInterpretedDataFrame[i] = 0x00;
			}
		}
		return mLastInterpretedDataFrame;
	}
	
	public void addData(byte[] data) {
		if(mAutoSetData) {
			setRawData(data);
		}
		addToHistory(data);
	}
	
	public void addToHistory(byte[] data) {
		if(mFrameNumberIndex >= 0 && mFrameNumberIndex < data.length - 1) {
			mLastReceivedFrameNumber = (int) (((0xFF & data[mFrameNumberIndex + 1]) << 8) | (0xFF & data[mFrameNumberIndex]));
		}
		synchronized(mDataHistory) {
			mDataHistory.addLast(new DataFrame(data));
			while(mDataHistory.size() > mMaxHistorySize && mDataHistory.size() > 0) {
				mDataHistory.removeFirst();
			}
		}
		notifyDataHistoryChanged();
	}
	
	public LinkedList<DataFrame> getDataHistory() {
		synchronized(mDataHistory) {
			return new LinkedList<DataFrame>(mDataHistory);
		}
	}
	
	public void clearHistory() {
		synchronized (mDataHistory) {
			mDataHistory.clear();
		}
	}
	
	public void setDataFormatter(int pos, DataFormatter formatter) {
		if(pos >= 0 && pos < mDataFormatter.length) {
			mDataFormatter[pos] = formatter;
			
			int bytesPerFrame = 0;
			for(int i = 0; i < mDataFormatter.length; ++i) {
				DataFormatter form = mDataFormatter[i];
				if(form != null) {
					bytesPerFrame += form.getNumberOfBytesPerData();
				}
			}
			mNumberOfBytesPerFrame = bytesPerFrame;
			
			notifyDataStructureChanged();
		}
	}
	
	public int getNumberOfBytesPerFrame() {
		return mNumberOfBytesPerFrame;
	}
	
	public DataFormatter[] getDataFormatters() {
//		TODO protect.
		return mDataFormatter;
	}
	
	public void setDataColor(int pos, Color color) {
		if(pos < mDataColors.length) {
			mDataColors[pos] = color;
			notifyDataStructureChanged();
		}
	}
	
	public Color[] getDataColors() {
		//TODO protect.
		return mDataColors;
	}
	
	public Object[] getCurrentData() {
		return mCurrentData;
	}
	
	public void setDimensions(int numberOfColumns, int numberOfRows) {
		
		if(mNumberOfColumns != numberOfColumns || mNumberOfRows != numberOfRows) {
			Object[] data = mCurrentData;
			DataFormatter[] formatters = mDataFormatter;
			Color[] colors = mDataColors;
			
			mNumberOfColumns = numberOfColumns;
			mNumberOfRows = numberOfRows;
			
			mCurrentData = new Object[mNumberOfColumns * mNumberOfRows];
			mDataFormatter = new DataFormatter[mNumberOfColumns * mNumberOfRows];
			mDataColors = new Color[mNumberOfColumns * mNumberOfRows];
			
			for(int i = 0; i < data.length && i < mCurrentData.length; ++i) {
				mCurrentData[i] = data[i];
			}
			for(int i = 0; i < formatters.length && i < mDataFormatter.length; ++i) {
				mDataFormatter[i] = formatters[i];
			}
			for(int i = 0; i < colors.length && i < mDataColors.length; ++i) {
				mDataColors[i] = colors[i];
			}
			
			notifyDataStructureChanged();
		}
	}
	
	
	public int getNumberOfColumns() {
		return mNumberOfColumns;
	}
	
	public int getNumberOfRows() {
		return mNumberOfRows;
	}
		
	public Hashtable<String, DataFormatter> getAvailableDataFormatters() {
		return new Hashtable<String, DataFormatter>(mAvailableDataFormatters);
	}
	
	public void addAvailableDataFormatter(DataFormatter formatter) {
		if(formatter == null) {
			return;
		}
		mAvailableDataFormatters.put(formatter.getName(), formatter);
	}
	
	public Hashtable<String, Color> getAvailableDataColors() {
		return new Hashtable<String, Color>(mAvailableDataColors);
	}
	
	public void addAvailableColor(String name, Color color) {
		if(name == null || color == null) {
			return;
		}
		mAvailableDataColors.put(name, color);
	}
	
	public synchronized boolean addDataListener(DataListener listener) {
		if(listener == null || mDataListeners.contains(listener)) {
			return false;
		}
		mDataListeners.add(listener);
		return true;
	}
	
	public synchronized boolean removeDataListener(DataListener listener) {
		if(listener == null || !mDataListeners.contains(listener)) {
			return false;
		}
		mDataListeners.remove(listener);
		return true;
	}
	
	public synchronized Vector<DataListener> getDataListeners() {
		return new Vector<DataListener>(mDataListeners);
	}
	
	public void setMaxHistorySize(int historySize) {
		mMaxHistorySize = historySize;
	}
	
	public int getMaxHistorySize() {
		return mMaxHistorySize;
	}
	
	public void readInitializationFile(String fileName) {
		//TODO read from file. If not found, initialize with specific Octavio object.
		
		if(fileName.equals("Default")) {
			OctavioDataStructure ods = new OctavioDataStructure();
			ods.setUpOctavioDataStructure(this);
		}
	}
	
	public void setSlotSizes(int masterSectionSize, int slot0SectionSize, int slotNSectionSize) {
		mMasterSectionSize = masterSectionSize;
		mSlot0SectionSize = slot0SectionSize;
		mSlotNSectionSize = slotNSectionSize;
		notifyDataStructureChanged();
	}
	
	public int getMasterSectionSize() {
		return mMasterSectionSize;
	}
	
	public int getSlot0SectionSize() {
		return mSlot0SectionSize;
	}
	
	public int getSlotNSectionSize() {
		return mSlotNSectionSize;
	}
	
	public boolean isAutoSettingData() {
		return mAutoSetData;
	}
	
	public void setAutoSetData(boolean autoSet) {
		mAutoSetData = autoSet;
	}
	
	public void setFrameNumberIndex(int index) {
		mFrameNumberIndex = index;
	}
	
	public int getFrameNumberIndex() {
		return mFrameNumberIndex;
	}
	
	public int getLastReceivedFrameNumber() {
		return mLastReceivedFrameNumber;
	}
	
	public int getLastInterpretedFrameNumber() {
		return mLastInterpretedFrameNumber;
	}
	
	public void setIgnoreFrameSynchronization(boolean ignore) {
		mIgnoreFrameSynchronization = ignore;
	}
	
	public boolean getIgnoreFrameSynchronization() {
		return mIgnoreFrameSynchronization;
	}
		
	
	public synchronized void notifyDataContentChanged() {
		for(Iterator<DataListener> i = mDataListeners.iterator(); i.hasNext();) {
			i.next().dataContentChanged();
		}
	}
	
	public synchronized void notifyDataStructureChanged() {
		for(Iterator<DataListener> i = mDataListeners.iterator(); i.hasNext();) {
			i.next().dataStructureChanged();
		}
	}
	
	public synchronized void notifyDataHistoryChanged() {
		for(Iterator<DataListener> i = mDataListeners.iterator(); i.hasNext();) {
			i.next().dataHistoryChanged();
		}
	}
	
	
	/** Protected ****************************************************************
	 *****************************************************************************/
	
		
		
	/** Private *******************************************************************
	 *****************************************************************************/
		
	


	/** Internal Classes *********************************************************
	 *****************************************************************************/





	/** Main Method **************************************************************
	 *****************************************************************************/
	 
	 

}
