/**
 * 
 * @author Christian Rempis
 * c.rempis@gmx.de
 * 
 */
package de.uos.ikw.octavio.spinalcordmanager.core;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Logger;

import de.uos.ikw.octavio.spinalcordmanager.core.DataManager;
import de.uos.ikw.octavio.spinalcordmanager.core.SystemObject;
import de.uos.ikw.octavio.spinalcordmanager.io.CommandWriter;
import de.uos.ikw.octavio.spinalcordmanager.io.ErrorStreamReader;
import de.uos.ikw.octavio.spinalcordmanager.io.SerialDataReader;

public class SpinalCordManager {
	
	
/** Member Variables *********************************************************
 *****************************************************************************/

	private static SpinalCordManager sInstance;
	private Vector<SystemObject> mSystemObjects = new Vector<SystemObject>();
	private DataManager mDataManager;
	private SerialDataReader mDataReader;
	private ErrorStreamReader mErrorReader;
	private CommandWriter mCommandWriter;
	private Thread mDataReaderThread;
	private Thread mErrorReaderThread;
	private Thread mCommandWriterThread;
	
/** Constructors *************************************************************
 *****************************************************************************/
	
	
	public SpinalCordManager() {
		
		mDataManager = new DataManager();
		addSystemObject(mDataManager);
	}
	
	
/** Public Static ************************************************************
 *****************************************************************************/
	
	public synchronized static SpinalCordManager getInstance() {
		if(sInstance == null) {
			sInstance = new SpinalCordManager();
		}
		return sInstance;
	}
	
/** Public *******************************************************************
 *****************************************************************************/
	
	public void setupManager() {
		mDataReader = new SerialDataReader(0, 0, 0, 0);
		mErrorReader = new ErrorStreamReader();
		mCommandWriter = new CommandWriter();
		
		mDataReaderThread = new Thread(mDataReader);
		mErrorReaderThread = new Thread(mErrorReader);
		mCommandWriterThread = new Thread(mCommandWriter);
		
		mDataReaderThread.start();
		mErrorReaderThread.start();
		mCommandWriterThread.start();
	}
	
	
	
	public void parseCommandLineArguments(String[] args) {
		
		
		mDataManager.readInitializationFile("Default");
	}
	
	
	public boolean addSystemObject(SystemObject object) {
		if(object == null || mSystemObjects.contains(object)) {
			return false;
		}
		mSystemObjects.add(object);
		return true;
	}
	
	public boolean removeSystemObject(SystemObject object) {
		if(object == null | !mSystemObjects.contains(object)) {
			return false;
		}
		mSystemObjects.remove(object);
		return true;
	}
	
	
	public DataManager getDataManager() {
		return mDataManager;
	}
	
	/**
	 * Initializes all registered SystemObjects.
	 * 
	 * @return true if all SystemObjects could be initialized successfully, otherwise false.
	 */
	public boolean init() {
		
		boolean ok = true;
		for(Iterator<SystemObject> i = mSystemObjects.iterator(); i.hasNext();) {
			SystemObject systemObject = i.next();
			if(!systemObject.init()) {
				Logger.getLogger("").warning("Could not initilialize SystemObject " + systemObject);
				ok = false;
			}
		}
		return ok;
	}
	
	/**
	 * Terminates all registered SystemObjects.
	 * 
	 * @return true if all SystemObjects could be terminated successfully, otherwise false.
	 */
	public boolean shutDown() {
		boolean ok = true;
		for(Iterator<SystemObject> i = mSystemObjects.iterator(); i.hasNext();) {
			SystemObject systemObject = i.next();
			if(!systemObject.shutDown()) {
				Logger.getLogger("").warning("Could not shut down SystemObject " + systemObject);
				ok = false;
			}
		}
		return ok;
	}
	
	public SerialDataReader getDataReader() {
		return mDataReader;
	}
	
	public ErrorStreamReader getErrorReader() {
		return mErrorReader;
	}
	
	public CommandWriter getCommandWriter() {
		return mCommandWriter;
	}
	
	
/** Private *******************************************************************
 *****************************************************************************/
	
	
	
	
	
/** Protected ****************************************************************
 *****************************************************************************/

	
	

/** Main Method **************************************************************
 *****************************************************************************/

	/**
	 * Main Application
	 */
	public static void main(String[] args) {
		SpinalCordManager scm = SpinalCordManager.getInstance();
		scm.parseCommandLineArguments(args);
		scm.init();
		scm.getDataManager().readInitializationFile("Default");
		scm.getDataManager().setLogFile("Logdata.txt");
		scm.getDataManager().setLogData(true);

		
		SerialDataReader reader = new SerialDataReader(9, 8, 32, 48);
		try {
			InputStream is = new FileInputStream(args[0]);
			reader.setInputStream(is, false);
			reader.run();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

}
