/**
 * 
 * @author Christian Rempis
 * c.rempis@gmx.de
 * 
 */
package de.uos.ikw.octavio.spinalcordmanager.io;


import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.LinkedList;
import java.util.Vector;

import de.uos.ikw.octavio.spinalcordmanager.core.SpinalCordManager;
import de.uos.ikw.octavio.spinalcordmanager.core.SystemObject;

public class CommandWriter implements Runnable, SystemObject {

	private PrintWriter mPrintWriter;
	private boolean mRunThread;
	private Vector<Command> mCommands = new Vector<Command>();
	private LinkedList<byte[]> mControlMessages = new LinkedList<byte[]>();
	private boolean mChangeAutoRepeat;
	private boolean mDesiredAutoRepeatOn;

	public static final String MODE_CHARACTER = "Mode Character"; // write data as single characters, followed by a return (win).
	public static final String MODE_BYTE = "Mode Byte"; //write data as a sequence of raw bytes (linux)
	
	public static final int NETWORK_TARGET_KERNEL = 1;
	public static final int NETWORK_TARGET_EEPROM = 2;
	
	private String mCommandOutputType   = MODE_CHARACTER; 
	private DataOutputStream mDataOutputStream = null;           // allows direct byte writing

	public CommandWriter() {
		SpinalCordManager.getInstance().addSystemObject(this);
		mDesiredAutoRepeatOn = false;
		mChangeAutoRepeat = false;
	}

	public void setOutputType(String type) {
		mCommandOutputType = type;
	}

	public void setOutputStream(OutputStream output) {
		if(mCommandOutputType == MODE_BYTE) {
			try {
				if(mDataOutputStream != null) {
					mDataOutputStream.close();
					mDataOutputStream = null;
				}
				if(output != null) {
					mDataOutputStream = new DataOutputStream(output);
				}
			}
			catch (IOException ioe){
			}
		}
		else {
			if(mPrintWriter != null) {
				mPrintWriter.close();
				mPrintWriter = null;
			}
			if(output != null) {
				mPrintWriter = new PrintWriter(output);
			}
		}

		try {
			synchronized(this) {
				notifyAll();
			}
		} 
		catch(IllegalMonitorStateException e) {
		}
		mRunThread = true;
	}


	public void run() {
		mRunThread = true;

		synchronized(this) {
			while(mDataOutputStream == null && mPrintWriter == null) {
				try {
					if(!mCommands.isEmpty()) {
						mCommands.clear();
					}
					
					wait();
				} catch (InterruptedException e) {
				} catch (IllegalMonitorStateException e) {
				}
			}
		}

		if(mCommandOutputType == MODE_BYTE)
		{

			while(mRunThread) {
				synchronized(this) {
					while(mDataOutputStream == null) {
						try {
							
							wait();
						} catch (InterruptedException e) {
						} catch (IllegalMonitorStateException e) {
						}
					}
				}

				try{
					while(!mControlMessages.isEmpty()) {
						byte[] message = null;
						synchronized(this) {
							message = mControlMessages.poll();
						}
						mDataOutputStream.write(message);
						mDataOutputStream.flush();
					}
					/*
					if(mChangeAutoRepeat) {
						mChangeAutoRepeat = false;
						mDataOutputStream.writeByte(0x72); // "r"
						if(mDesiredAutoRepeatOn) {
							mDataOutputStream.writeByte(0x31); //"1"
						}
						else {
							mDataOutputStream.writeByte(0x30); //"0"
						}
						mDataOutputStream.flush();
					}
					if(mChangeNetworkMode) {
						mChangeNetworkMode = false;
						mDataOutputStream.writeByte(0x6e);; // "n"
						if(mDesiredNetworkMode) {
							mDataOutputStream.writeByte(0x31); //"1"
						}
						else {
							mDataOutputStream.writeByte(0x30); //"0"
						}
						mDataOutputStream.flush();
					}
					*/
					if(!mCommands.isEmpty()) {
						Command command = mCommands.remove(0);
						if(command.isSendInDeploymentMode()) {
							mDataOutputStream.writeByte(0x64); //'d'
							mDataOutputStream.writeByte((byte) command.getData().length);
						}
						else {
							mDataOutputStream.writeByte(0x63); // "c"
							if(command.isDiscardedIfNotSendImmediately()) {
								mDataOutputStream.writeByte(0x01);
							}
							else {
								mDataOutputStream.writeByte(0x00);
							}
							//2 bytes for expiration frame number.
							mDataOutputStream.writeByte((byte) command.getExpirationFrameNumber());
							mDataOutputStream.writeByte((byte) (command.getExpirationFrameNumber() >> 8));
						}
						mDataOutputStream.flush();
						int[] data = command.getData();

						for(int i = 0; i < data.length; ++i) {
							mDataOutputStream.writeByte((byte) data[i]);
							//out.flush();
						}
						mDataOutputStream.flush();
					}
					else {
						synchronized(this) {
							try {
								wait();
							} catch (InterruptedException e) {
							}
						}
					}
				}
				catch(IOException ioe) {
				}
			}
		}
		else // MODE_CHARACTER
		{
			//mPrintWriter.println("");
			//mPrintWriter.flush();
			while(mRunThread) {
				synchronized(this) {
					while(mPrintWriter == null) {
						try {
							
							wait();
						} catch (InterruptedException e) {
						} catch (IllegalMonitorStateException e) {
						}
					}
				}

				if(mChangeAutoRepeat) {
					mChangeAutoRepeat = false;
					mPrintWriter.println("r");
					if(mDesiredAutoRepeatOn) {
						mPrintWriter.println("1");
					}
					else {
						mPrintWriter.println("0");
					}
					mPrintWriter.flush();
				}
				if(!mCommands.isEmpty()) {
					Command command = mCommands.remove(0);
					mPrintWriter.println("c");
					if(command.isDiscardedIfNotSendImmediately()) {
						mPrintWriter.println("1");
					}
					else {
						mPrintWriter.println("0");
					}
					mPrintWriter.println("" + command.getExpirationFrameNumber());

					mPrintWriter.flush();
					int[] data = command.getData();

					for(int i = 0; i < data.length; ++i) {

						mPrintWriter.println(data[i]);
						//mPrintWriter.flush();
					}
					mPrintWriter.flush();
				}
				else {
					try {
						synchronized(this) {
							wait();
						}
					} catch (InterruptedException e) {
					}
				}
			}

			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
			}
		}
	}

	public void stopThread() {
		mRunThread = false;
	}

	public synchronized void sendCommand(int[] command, boolean verifySending, 
			boolean discardIfNotSendImmediately, int expirationFrameNumber, boolean deploymentMode) 
	{

		mCommands.add(new Command(command, verifySending, 
				discardIfNotSendImmediately, expirationFrameNumber,  deploymentMode));
		if(false) {
			for(int i = 0; i < command.length; ++i) {
				System.out.print(command[i] + ", ");
			}
			System.out.println("");
		}
		notifyAll();
	}
	
	public synchronized void sendMessage(byte[] message) {
		mControlMessages.addLast(message.clone());
		notifyAll();
	}

	public synchronized void setAutoRepeat(boolean activate) {
		/*
		mChangeAutoRepeat = true;
		mDesiredAutoRepeatOn = activate;
		notifyAll();
		*/
		byte[] message = new byte[2];
		message[0] = 0x72; //r
		if(activate) {
			message[1] = 0x31; //1 
		}
		else {
			message[1] = 0x30; //0
		}
		sendMessage(message);
	}

	public synchronized void setVerbose(boolean verbose) {
		byte[] message = new byte[2];
		message[0] = 0x56; //V
		if(verbose) {
			message[1] = 0x31; //1 
		}
		else {
			message[1] = 0x30; //0
		}
		sendMessage(message);
	}
	
	public synchronized void setNetworkMode(boolean activate, int target) {
		/*
		mChangeNetworkMode = true;
		mDesiredNetworkMode = activate;
		notifyAll();
		*/
		
		if(target == NETWORK_TARGET_KERNEL) {
			byte[] message = new byte[2];
			message[0] = 0x6e; //n
			if(activate) {
				message[1] = 0x31; //1 
			}
			else {
				message[1] = 0x30; //0
			}
			sendMessage(message);
		}
		else {
			int[] message = new int[4];
			message[0] = Command.nextCommandId();
			message[1] = 255;
			if(activate) {
				message[2] = 33; //1 
			}
			else {
				message[2] = 34; //0
			}
			message[3] = 0;
			for(int i = 0; i < 2; ++i) { //TODO check why command has to be sent twice
				sendCommand(message, false,	false, -1, false);
			}
		}
	}

	public boolean init() {
		return true;
	}

	public boolean shutDown() {
		stopThread();
		synchronized(this) {
			notifyAll();
		}
		return true;
	}

}
