/**
 * 
 * @author Christian Rempis
 * c.rempis@gmx.de
 * 
 */
package de.uos.ikw.octavio.spinalcordmanager.io;


import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Vector;

import de.uos.ikw.octavio.spinalcordmanager.core.SpinalCordManager;
import de.uos.ikw.octavio.spinalcordmanager.core.SystemObject;


public class ErrorStreamReader implements Runnable, SystemObject {

	private InputStream mSerialLine;
	private boolean mRunThread;
	private Vector<ErrorStreamListener> mErrorStreamListeners = new Vector<ErrorStreamListener>();
	private byte[] mInputBuffer = new byte[1024];
	private int mPosition;
	
	public ErrorStreamReader() {
		mRunThread = true;
		SpinalCordManager.getInstance().addSystemObject(this);
	}
	
	public void stopReader() {
		mRunThread = false;
	}
	
	public void setSerialLine(InputStream input) {
		mSerialLine = input;
	}
	
	public boolean addErrorStreamListener(ErrorStreamListener listener) {
		if(listener == null || mErrorStreamListeners.contains(listener)) {
			return false;
		}
		mErrorStreamListeners.add(listener);
		return true;
	}
	
	public boolean removeErrorStreamListener(ErrorStreamListener listener) {
		if(listener == null || !mErrorStreamListeners.contains(listener)) {
			return false;
		}
		mErrorStreamListeners.remove(listener);
		return true;
	}
	
	public Vector<ErrorStreamListener> getErrorStreamListeners() {
		return new Vector<ErrorStreamListener>(mErrorStreamListeners);
	}
	
	public void reportErrorMessageReceived(String message) {
		for(Iterator<ErrorStreamListener> i = mErrorStreamListeners.iterator(); i.hasNext();) {
			i.next().receivedErrorMessage(message);
		}
	}
	
	public void reportDataMessageReceived(String messageHeader, byte[] data) {
		for(Iterator<ErrorStreamListener> i = mErrorStreamListeners.iterator(); i.hasNext();) {
			i.next().receivedDataMessage(messageHeader, data);
		}
	}
	

	public void run() {
		mRunThread = true;
		
		mPosition = 0;
		clearBuffer();
		
		while(mRunThread) {
			while(mSerialLine == null) {
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {
				}
			}
			
			try {
				mInputBuffer[mPosition] = (byte) mSerialLine.read();
				if(mInputBuffer[mPosition] == '\n') {
					processMessage(mInputBuffer, mPosition);
				}
				else {
					mPosition++;
					if(mPosition >= mInputBuffer.length) {
						mPosition = mInputBuffer.length - 1;
					}
				}
				
			} catch (IOException e) {
				e.printStackTrace();
			}
			
		}
	}

	public boolean init() {
		return true;
	}

	public boolean shutDown() {
		stopReader();
		return true;
	}
	
	private void clearBuffer() {
		for(int i = 0; i < mInputBuffer.length; ++i) {
			mInputBuffer[i] = 0x00;
		}
		mPosition = 0;
	}
	
	private void processMessage(byte[] msg, int length) {
		String message = new String(msg).substring(0, length);
		
		//System.out.println("Received: " + message);
		
		if(message.startsWith("[B]")) {
			int size = 0;
			try {
				size = (int) mSerialLine.read();
			} catch (IOException e1) {
			}
			//System.out.println("Receiving bytes: " + size);
			byte[] data = new byte[size];
			int readBytes = 0;
			while(readBytes < (size) && mRunThread) {
				
				try {
					readBytes += mSerialLine.read(data, readBytes, size - readBytes);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			if(!mRunThread) {
				return;
			}
			reportDataMessageReceived(message.substring(3, Math.min(message.indexOf("]", 3) + 1, message.length())), data);
		}
		else {
			reportErrorMessageReceived(message);
		}
		clearBuffer();
	}
	
	
	
}
