/**
 * 
 * @author Christian Rempis
 * c.rempis@gmx.de
 * 
 */
package de.uos.ikw.octavio.spinalcordmanager.io;

import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;

import de.uos.ikw.octavio.spinalcordmanager.core.DataListener;
import de.uos.ikw.octavio.spinalcordmanager.core.DataManager;
import de.uos.ikw.octavio.spinalcordmanager.core.SpinalCordManager;
import de.uos.ikw.octavio.spinalcordmanager.core.SystemObject;

public class SerialDataReader implements DataListener, Runnable, SystemObject {

	private InputStream mInputStream;
	private boolean mRunReader;
	private int mNumberOfSlots;
	private int mSlotSize;
	private int mSlot0Size;
	private int mMasterSectionSize;
	private boolean mReadNextFrame;
	private boolean mSingleFrameMode;
	private boolean mEnableSingleFrameMode;
	private byte[] mFrame;
	private boolean[] mEnabledSlots;
	private int mNumberOfSuccessivelyReadFrames;
	private DataManager mDataManager;
	private Vector<RawByteListener> mRawByteListeners = new Vector<RawByteListener>();

	public SerialDataReader(int numberOfSlots, int masterSectionSize, int slot0Size, int slotSize) {
		mInputStream = null;
		mNumberOfSlots = numberOfSlots;
		mSlotSize = slotSize;
		mSlot0Size = slot0Size;
		mMasterSectionSize = masterSectionSize;
		
		mFrame = new byte[SpinalCordManager.getInstance().getDataManager().getNumberOfBytesPerFrame()];
		mEnabledSlots = new boolean[SpinalCordManager.getInstance().getDataManager().getNumberOfBytesPerFrame()];
		SpinalCordManager.getInstance().getDataManager().addDataListener(this);
		SpinalCordManager.getInstance().addSystemObject(this);
		mDataManager = SpinalCordManager.getInstance().getDataManager();
	}
	
	public void setInputStream(InputStream input, boolean enableSingleFrameMode) {
		mInputStream = input;
		mEnableSingleFrameMode = enableSingleFrameMode;
		synchronized(this) {
			notifyAll();
		}
	}
	
	public boolean setSingleFrameMode(boolean activateSingleFrameMode) {
		if(mEnableSingleFrameMode) {
			mSingleFrameMode = activateSingleFrameMode;
			if(mSingleFrameMode) {
				mReadNextFrame = true;
			}
			else {
				mReadNextFrame = false;
			}
		}
		else {
			mSingleFrameMode = false;
		}
		return mSingleFrameMode;
	}
	

	public void stopReader() {
		mRunReader = false; 
	}
	
	public void readNextFrame() {
		if(mSingleFrameMode) {
			mReadNextFrame = true;
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		
		//System.out.println("Starting to read...");
		
		mDataManager = SpinalCordManager.getInstance().getDataManager();
		
		mRunReader = true;
		
		synchronizeWithBus();

		byte[] data;
					
		int index = 0;
		boolean foundFirstMasterSection = false;
		mNumberOfSuccessivelyReadFrames = 0;
		
		try {
			while(mRunReader) {
				
				//System.out.println("Before while(mInputStream == null)." + (mInputStream));
				while(mInputStream == null) {
					try {
						synchronized(this) {
							wait();
						}
					} catch (InterruptedException e) {
					}
				}
				
				
				//System.out.println("Before blocking read of byte");
				data = readNBytes(1);
				
				if(index >= mFrame.length)	{
					synchronizeWithBus();
					index = 1;
					continue;
				}

				//System.out.println("Read byte " + data[0]);
				mFrame[index] = data[0];
				index++;
				
				//read raw bytes instead of real frames (for debugging purposes).
				if(mDataManager.getIgnoreFrameSynchronization()) {
					while(mRunReader && mInputStream != null && mDataManager.getIgnoreFrameSynchronization()) {
						data = readNBytes(mFrame);
						updateDataTable(mFrame);
					}
					index = 0;
					foundFirstMasterSection = false;
					mNumberOfSuccessivelyReadFrames = 0;
					continue;
				}
			
				if(data[0] == 0x55) {
					//detected the master section of the frame.
					
					if(foundFirstMasterSection) {
						processFrame(mFrame, mEnabledSlots);
					}
					
					foundFirstMasterSection = true;
					
					mFrame[0] = 0x55;
					index = 1;
					for(int i = 0; i < mEnabledSlots.length; ++i) {
						mEnabledSlots[i] = false;
					}
					
					//read the rest of the header
					//data = readNBytes(7);
					data = readNBytes(mMasterSectionSize - 1);
					for(int i = 0; i < data.length && index < mFrame.length; ++i) {
						mFrame[index] = data[i];
						index++;
					}
			
					if(data[mMasterSectionSize - 2] != 0x55) {
						foundFirstMasterSection = false;
						//this byte has to be 0x55, otherwise we have to resynchronize.
						System.out.println("Out of sync: Header " + mNumberOfSuccessivelyReadFrames);
						mNumberOfSuccessivelyReadFrames = 0;
						if(!synchronizeWithBus()) {
							//return;
						}
						index = 1;
						continue;
					}
				}
				else if(data[0] == 0x00) {
					//detected special slot 0.
					mEnabledSlots[0] = true;
					
					//read the rest of the slot.
					data = readNBytes(mSlot0Size - 1);
					for(int i = 0; i < data.length && index < mFrame.length; ++i) {
						mFrame[index] = data[i];
						index++;
					}
			
					//last byte has to be a sync byte 0x55!
					if(data[data.length - 1] != 0x55) {
						System.out.println("Out of sync: Special" + mNumberOfSuccessivelyReadFrames);
						mNumberOfSuccessivelyReadFrames = 0;
						foundFirstMasterSection = false;
						if(!synchronizeWithBus()) {
							//return;
						}
						index = 1;
						continue;
					}
				}
				else {
					//detected one of the normal slots.
					int slotNo =  (0xFF & (int) data[0]);
					if(slotNo > 0 && slotNo < mEnabledSlots.length) {
						mEnabledSlots[slotNo] = true;
					}
					
					
					data = readNBytes(mSlotSize - 1);
					for(int i = 0; i < data.length && index < mFrame.length; ++i) {
						mFrame[index] = data[i];
						index++;
					}
			
					//last byte has to be a sync byte 0x55!
					if(data[data.length - 1] != 0x55) {
						System.out.println("Out of sync: Normal Slot" + mNumberOfSuccessivelyReadFrames);
						mNumberOfSuccessivelyReadFrames = 0;
						foundFirstMasterSection = false;
						if(!synchronizeWithBus()) {
							//return;
						}
						index = 1;
						continue;
					}
				}
				 mNumberOfSuccessivelyReadFrames++;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}	
		//System.out.println("End of run().");
	}
	
	public void processFrame(byte[] data, boolean[] enabledSlots) {
		byte[] textData = new byte[mMasterSectionSize + mSlot0Size + (mNumberOfSlots - 1) * mSlotSize];
		if(textData.length > 0) {
			textData[0] = 0x55;
			int index = 1;
			int dataIndex = 0;
			for(int i = 0; i < mMasterSectionSize - 1; ++i) {
				textData[index] = data[i];
				index++;
				dataIndex++;
			}
	//		for(int i = 0; i < 8; ++i) {
	//			textData[index] = 0x00;
	//			index++;
	//		}
			if(enabledSlots[0]) {
				for(int i = 0; i < mSlot0Size; ++i) {
					textData[index] = data[dataIndex];
					index++;
					dataIndex++;
				}
			}
			else {
				for(int i = 0; i < mSlot0Size; ++i) {
					textData[index] = 0x00;
					index++;
				}
			}
			for(int slot = 1; slot < mNumberOfSlots; ++slot) {
				if(enabledSlots[slot]) {
					for(int i = 0; i < mSlotSize; ++i) {
						textData[index] = data[dataIndex];
						index++;
						dataIndex++;
					}
				}
				else {
					for(int i = 0; i < mSlotSize; ++i) {
						textData[index] = 0x00;
						index++;
					}
				}
			}
			updateDataTable(textData);
			
			while(mEnableSingleFrameMode && mSingleFrameMode && !mReadNextFrame) {
				try {
					Thread.sleep(0, 500);
				} catch (InterruptedException e) {
				}
			}
			mReadNextFrame = false;
		}
		
		//printData(textData);
	}
	
	private void updateDataTable(byte[] data) {
		//SpinalCordManager.getInstance().getDataManager().setRawData(data);
		//SpinalCordManager.getInstance().getDataManager().addToHistory(data);
		mDataManager.addData(data);
	}
	
//	private void printData(byte[] data) {
//		byte b = 0x55;
//		System.out.println("Stopp is: " + Byte.toString(b));
//		for(int i = 0; i < mSlotSize / 2; ++i) {
//			for(int j = 0; j < mNumberOfSlots; ++j) {
//				
//				int value = ((0xFF & (int) data[mSlotSize * j + i * 2 + 1]) << 8 | (0xFF & (int) data[mSlotSize * j + i * 2]));
//
//				System.out.print("" + value + "\t\t");
////				System.out.print(Byte.toString(data[mSlotSize * j * 2 + i * 2 + 1]) + "," 
////				   + Byte.toString(data[mSlotSize * j * 2 + i * 2 + 0]) + "\t");
//			}
//			System.out.println("");
//		}
//		for(int i = 0; i < data.length; ++i) {
//			System.out.print("" + data[i] + ",");
//		}
//		
//		System.out.println("\n---");
//	}
	
	private byte[] readNBytes(int size) throws IOException  {
		return readNBytes(new byte[size]);
	}
	
	private byte[] readNBytes(byte[] data) throws IOException {

		int readDataSize = 0;
		
		byte[] input = new byte[1];
		while(readDataSize < data.length && mInputStream != null) {
			try {
				readDataSize += mInputStream.read(input, 0, 1);
				
				//TODO do we really need this (duplication could also be removed in the kernel module)
				//TODO does this still work with the WIN version, as there is no duplicated 0x0d?
				//mask 0x0D
				
//				if(input[0] == 0x0D) {
//					//System.out.println("0x0D");
//					while(mInputStream.read(input, 0, 1) == 0) {
//						//System.out.println("while mInputStream.read()");;
//					}
//				}
				
				if(readDataSize > 0) {
					data[readDataSize - 1] = input[0];
					notifyRawByteListeners(input[0]);
				}
				
				
			} catch (IndexOutOfBoundsException e) {
				e.printStackTrace();
				return new byte[0];
			}
		}
		
		return data;
	}
	
	//TODO check with variable master section size
	private boolean synchronizeWithBus() {
		
		int masterId = -1;
		int secureCounter = 100000; //makes sure that we cancel synchronization after a while
	
		System.out.println("SYNC");
		
		byte[] input;

		
		try {
			while(secureCounter > 0) {
				secureCounter--;
				
				if(mDataManager.getIgnoreFrameSynchronization()) {
					//do not synchronize if in ignore mode.
					return true;
				}
			
				input = readNBytes(1);
				
				//System.out.println("Secure counter = " + secureCounter);

				if(input[0] == 0x55) {
			
					//System.out.println("1, secure counter = " + secureCounter);
					input = readNBytes(1);

					if(input[0] == 0x55) {
						//System.out.println("2");
						readNBytes(4);
						input = readNBytes(5);
						
						masterId = input[0];

						if(input[2] == 0x55) {
							//System.out.println("3");
							if(input[4] != masterId) {
								//continue; //enable if this additional check should be performed.
							}

							if(input[3] == 0x00) {
								//System.out.println("4");
			
								//special frame 0 detected 
								input = readNBytes(mSlot0Size - 2);

								if(input[mSlot0Size - 3] == 0x55) {
									//System.out.println("5");

									//synchronized
									return true;
								}
							}
							else if(input[3] > 0x00 && input[3] < 0x10) {
								//System.out.println("6, input[3] = " + (int)input[3]);

								//found normal slot
								input = readNBytes(mSlotSize - 2);
			                    
								//System.out.println("-4 = " + (int)input[mSlotSize - 4] + ", -3 = " + (int)input[mSlotSize - 3]);
								if(input[mSlotSize - 3] == 0x55) {
									//System.out.println("7");

									//synchronized
									return true;
								}
							}
						}
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		System.out.println("Could not synchronize with spinal cord.");
		return false;
	}

	public void dataContentChanged() {
	}

	public void dataStructureChanged() {
		DataManager dm = SpinalCordManager.getInstance().getDataManager();
		mMasterSectionSize = dm.getMasterSectionSize();
		mNumberOfSlots = dm.getNumberOfColumns();
		mSlot0Size = dm.getSlot0SectionSize();
		mSlotSize = dm.getSlotNSectionSize();
		mFrame = new byte[SpinalCordManager.getInstance().getDataManager().getNumberOfBytesPerFrame()];
		mEnabledSlots = new boolean[SpinalCordManager.getInstance().getDataManager().getNumberOfBytesPerFrame()];
		
	}
	
	public void dataHistoryChanged() {
	}
	
	public boolean init() {
		return true;
	}

	public boolean shutDown() {
		stopReader();
		return true;
	}
	
	public void addRawByteListener(RawByteListener listener) {
		if(listener != null && !mRawByteListeners.contains(listener)) {
			mRawByteListeners.add(listener);
		}
	}
	
	public void removeRawByteListener(RawByteListener listener) {
		if(listener != null && mRawByteListeners.contains(listener)) {
			mRawByteListeners.remove(listener);
		}
	}
	
	public void notifyRawByteListeners(byte receivedByte) {
		for(int i = 0; i < mRawByteListeners.size(); ++i) {
			mRawByteListeners.elementAt(i).byteReceived(receivedByte);
		}
	}
	
	public Vector<RawByteListener> getRawByteListeners() {
		return mRawByteListeners;
	}
	
}
