/**
 * 
 * @author Christian Rempis
 * c.rempis@gmx.de
 * 
 */
package de.uos.ikw.octavio.spinalcordmanager.test;

import java.awt.Color;

import de.uos.ikw.octavio.spinalcordmanager.core.DataListener;
import de.uos.ikw.octavio.spinalcordmanager.core.DataManager;
import de.uos.ikw.octavio.spinalcordmanager.formatter.DataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.Int16DataFormatter;
import de.uos.ikw.octavio.spinalcordmanager.formatter.Int8DataFormatter;
import junit.framework.TestCase;

public class TestDataManager extends TestCase {

	protected void setUp() throws Exception {
		super.setUp();
	}

	protected void tearDown() throws Exception {
		super.tearDown();
	}

	
	public void testConstructor() {
		DataManager dm = new DataManager();
		
		assertNotNull(dm.getDataFormatters());
		assertEquals(dm.getDataFormatters().length, 0);
		assertNotNull(dm.getDataColors());
		assertEquals(dm.getDataColors().length, 0);
		assertNotNull(dm.getCurrentData());
		assertEquals(dm.getCurrentData().length, 0);
		assertEquals(dm.getNumberOfColumns(), 0);
		assertEquals(dm.getNumberOfRows(), 0);
		assertEquals(dm.getMaxHistorySize(), 10);
		assertEquals(dm.getDataListeners().size(), 0);
		assertEquals(dm.getDataHistory().size(), 0);
		
		DataTestListener listener = new DataTestListener();
		assertTrue(dm.addDataListener(listener));
		assertEquals(listener.mCountStructureChanged, 0);
		
		//test dimension
		dm.setDimensions(5, 10);
		assertEquals(dm.getDataFormatters().length, 5 * 10);
		assertEquals(dm.getDataColors().length, 5 * 10);
		assertEquals(dm.getCurrentData().length, 5 * 10);
		assertEquals(dm.getNumberOfColumns(), 5);
		assertEquals(dm.getNumberOfRows(), 10);
		assertEquals(listener.mCountStructureChanged, 1);
		
		dm.setDimensions(7, 99);
		assertEquals(dm.getDataFormatters().length, 7 * 99);
		assertEquals(dm.getDataColors().length, 7 * 99);
		assertEquals(dm.getCurrentData().length, 7 * 99);
		assertEquals(dm.getNumberOfColumns(), 7);
		assertEquals(dm.getNumberOfRows(), 99);
		assertEquals(listener.mCountStructureChanged, 2);
		
		//TODO check if sizes are important.
		assertTrue(dm.getAvailableDataColors().size() > 10);
		assertTrue(dm.getAvailableDataFormatters().size() > 5);
		
	}
	
	
	public void testDataHistory() {
		DataManager dm = new DataManager();
		
		dm.setMaxHistorySize(3);
		assertEquals(dm.getMaxHistorySize(), 3);
		
		byte[] data = new byte[10];
		data[0] = 1;
		dm.addToHistory(data);
		
		assertEquals(dm.getDataHistory().size(), 1);
		assertEquals(dm.getDataHistory().getFirst().getData()[0], 1);
		
		data[0] = 2;
		dm.addToHistory(data);
		
		assertEquals(dm.getDataHistory().size(), 2);
		assertEquals(dm.getDataHistory().getFirst().getData()[0], 2);
		assertEquals(dm.getDataHistory().getLast().getData()[0], 1);
		
		data[0] = 3;
		dm.addToHistory(data);
		
		assertEquals(dm.getDataHistory().size(), 3);
		assertEquals(dm.getDataHistory().getFirst().getData()[0], 3);
		assertEquals(dm.getDataHistory().getLast().getData()[0], 1);
		
		data[0] = 4;
		dm.addToHistory(data);
		
		assertEquals(dm.getDataHistory().size(), 3);
		assertEquals(dm.getDataHistory().getFirst().getData()[0], 4);
		assertEquals(dm.getDataHistory().getLast().getData()[0], 2);
		
		dm.clearHistory();
		assertEquals(dm.getDataHistory().size(), 0);
	}
	
	public void testAddAndRemoveDataListeners() {
		
		DataManager dm = new DataManager();
		DataTestListener listener1 = new DataTestListener();
		DataTestListener listener2 = new DataTestListener();
		
		
		//Adding Listeners
		assertEquals(dm.getDataListeners().size(), 0);
		
		assertFalse(dm.addDataListener(null));
		assertTrue(dm.addDataListener(listener1));
		assertEquals(dm.getDataListeners().size(), 1);
		
		assertFalse(dm.addDataListener(listener1));
		assertEquals(dm.getDataListeners().size(), 1);
		
		assertTrue(dm.addDataListener(listener2));
		assertEquals(dm.getDataListeners().size(), 2);
		
		assertFalse(dm.removeDataListener(null));
		assertEquals(dm.getDataListeners().size(), 2);
		
		
		//listener notification
		
		assertEquals(listener1.mCountContentChanged, 0);
		assertEquals(listener1.mCountStructureChanged, 0);
		assertEquals(listener2.mCountContentChanged, 0);
		assertEquals(listener2.mCountStructureChanged, 0);
		
		dm.setDimensions(1, 1);
		
		assertEquals(listener1.mCountContentChanged, 0);
		assertEquals(listener1.mCountStructureChanged, 1);
		assertEquals(listener2.mCountContentChanged, 0);
		assertEquals(listener2.mCountStructureChanged, 1);
		
		
		byte[] data = new byte[0];
		assertTrue(dm.setRawData(data));
		
		assertEquals(listener1.mCountContentChanged, 1);
		assertEquals(listener1.mCountStructureChanged, 1);
		assertEquals(listener2.mCountContentChanged, 1);
		assertEquals(listener2.mCountStructureChanged, 1);
		
		dm.setDataColor(0, Color.red);
		assertEquals(listener2.mCountStructureChanged, 2);
		assertEquals(listener1.mCountStructureChanged, 2);
		
		dm.setDataFormatter(0, new DataFormatter("", 0));
		
		assertEquals(listener2.mCountStructureChanged, 3);
		assertEquals(listener1.mCountStructureChanged, 3);
		
		
		//listener removal
		
		assertTrue(dm.removeDataListener(listener1));
		assertEquals(dm.getDataListeners().size(), 1);
		assertSame(dm.getDataListeners().firstElement(), listener2);
		
		assertFalse(dm.removeDataListener(listener1));
		assertEquals(dm.getDataListeners().size(), 1);
		
		assertTrue(dm.removeDataListener(listener2));
		assertEquals(dm.getDataListeners().size(), 0);
		
	}
	
	
	public void testSetDataFormatters() {
		DataManager dm = new DataManager();
		
		dm.setDimensions(2, 3);
		
		DataFormatter[] formatters = dm.getDataFormatters();
		assertEquals(formatters.length, 6);
		for(int i = 0; i < formatters.length; ++i) {
			assertNull(formatters[i]);
		}
		
		DataFormatter f1 = new DataFormatter("F1", 0);
		DataFormatter f2 = new DataFormatter("F2", 0);
		DataFormatter f3 = new DataFormatter("F3", 0);
		DataFormatter f4 = new DataFormatter("F4", 0);
		
		dm.setDataFormatter(0, f1);
		dm.setDataFormatter(1, f2);
		dm.setDataFormatter(2, f2);
		dm.setDataFormatter(3, f3);
		dm.setDataFormatter(4, f4);
		dm.setDataFormatter(5, f4);
		
		formatters = dm.getDataFormatters();
		assertSame(formatters[0], f1);
		assertSame(formatters[1], f2);
		assertSame(formatters[2], f2);
		assertSame(formatters[3], f3);
		assertSame(formatters[4], f4);
		assertSame(formatters[5], f4);
		
		dm.setDimensions(3, 3);
		formatters = dm.getDataFormatters();
		assertEquals(formatters.length, 9);
		
		assertSame(formatters[0], f1);
		assertSame(formatters[1], f2);
		assertSame(formatters[2], f2);
		assertSame(formatters[3], f3);
		assertSame(formatters[4], f4);
		assertSame(formatters[5], f4);
		assertNull(formatters[6]);
		assertNull(formatters[7]);
		assertNull(formatters[8]);
		
		dm.setDimensions(2, 2);
		formatters = dm.getDataFormatters();
		assertEquals(formatters.length, 4);
		
		assertSame(formatters[0], f1);
		assertSame(formatters[1], f2);
		assertSame(formatters[2], f2);
		assertSame(formatters[3], f3);
	}
	
	public void testSetDataColors() {
		DataManager dm = new DataManager();
		
		dm.setDimensions(2, 3);
		
		Color[] colors = dm.getDataColors();
		assertEquals(colors.length, 6);
		for(int i = 0; i < colors.length; ++i) {
			assertNull(colors[i]);
		}
		
		Color red = Color.red;
		Color green = Color.green;
		Color cyan = Color.cyan;
		Color black = Color.black;
		
		dm.setDataColor(0, red);
		dm.setDataColor(1, green);
		dm.setDataColor(2, green);
		dm.setDataColor(3, cyan);
		dm.setDataColor(4, black);
		dm.setDataColor(5, black);
		
		colors = dm.getDataColors();
		assertSame(colors[0], red);
		assertSame(colors[1], green);
		assertSame(colors[2], green);
		assertSame(colors[3], cyan);
		assertSame(colors[4], black);
		assertSame(colors[5], black);
		
		dm.setDimensions(3, 3);
		colors = dm.getDataColors();
		assertEquals(colors.length, 9);
		
		assertSame(colors[0], red);
		assertSame(colors[1], green);
		assertSame(colors[2], green);
		assertSame(colors[3], cyan);
		assertSame(colors[4], black);
		assertSame(colors[5], black);
		assertNull(colors[6]);
		assertNull(colors[7]);
		assertNull(colors[8]);
		
		dm.setDimensions(2, 2);
		colors = dm.getDataColors();
		assertEquals(colors.length, 4);
		
		assertSame(colors[0], red);
		assertSame(colors[1], green);
		assertSame(colors[2], green);
		assertSame(colors[3], cyan);
	}
	
	
	public void testSetRawData() {
		
		DataManager dm = new DataManager();
		Int8DataFormatter int8 = new Int8DataFormatter();
		Int16DataFormatter int16 = new Int16DataFormatter();
		
		dm.setDimensions(1, 2);
		assertEquals(dm.getCurrentData().length, 2);
		dm.setDataFormatter(0, int8);
		dm.setDataFormatter(1, int16);
		
		assertEquals(dm.getNumberOfBytesPerFrame(), 3);
		
		{
			byte[] data = {0x55, 0x00};
			assertFalse(dm.setRawData(data));
			assertNull(dm.getCurrentData()[0]);
		}
		
		{
			byte[] data = {0x55, 0x00, 0x10, 0x19};
			assertFalse(dm.setRawData(data));
			assertNull(dm.getCurrentData()[0]);
		}
		
		{
			byte[] data = {0x55, 0x00, 0x01};
			assertTrue(dm.setRawData(data));
			assertNotNull(dm.getCurrentData()[0]);
			assertNotNull(dm.getCurrentData()[1]);
			assertEquals(((Integer) dm.getCurrentData()[0]).intValue(), 85);
		}
		dm.setDataFormatter(0, null);
		assertEquals(dm.getNumberOfBytesPerFrame(), 2);
		
		{
			byte[] data = {0x55, 0x00, 0x10};
			assertFalse(dm.setRawData(data));
			assertNotNull(dm.getCurrentData()[0]);
			assertEquals(((Integer) dm.getCurrentData()[0]).intValue(), 85);
			assertEquals(((Integer) dm.getCurrentData()[1]).intValue(), 256);
		}
		
		{
			byte[] data = {0x01, 0x10};
			assertTrue(dm.setRawData(data));
			assertNotNull(dm.getCurrentData()[0]);
			assertEquals(((Integer) dm.getCurrentData()[0]).intValue(), 85);
			assertEquals(((Integer) dm.getCurrentData()[1]).intValue(), 4097);
		}
		
		
		
	}
	
	public class DataTestListener implements DataListener {

		public int mCountContentChanged;
		public int mCountStructureChanged;
		public int mCountHistoryChanged;
		
		public void dataContentChanged() {
			mCountContentChanged++;
		}

		public void dataStructureChanged() {
			mCountStructureChanged++;
		}
		
		public void dataHistoryChanged() {
			mCountHistoryChanged++;
		}
		
		
	}
	
}
